  
"
    __________________________________________________
    @Alejandro F. Reimondo [aleReimondo@smalltalking.net] 11/10/99
    This proyect sources are property of : Alejandro F. Reimondo
    Its use is NOT legal without written rights.
    Amenabar 3476 Piso: 2 Dto: B.
    Ciudad de Buenos Aires.
    Argentina.
    Tel: (54)-1-703-5586.
    __________________________________________________
"!

ARObject subclass: #SkinDetectionObject
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

SkinDetectionObject subclass: #ColorDetector
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: ''    !

ColorDetector subclass: #ColorDetectorComposite
  instanceVariableNames: 
    ' contents '
  classVariableNames: ''
  poolDictionaries: '' !

ColorDetector subclass: #ColorPredicateDetector
  instanceVariableNames: 
    ' colorPredicate '
  classVariableNames: ''
  poolDictionaries: ''   !

ColorPredicateDetector class instanceVariableNames: ' colorPredicate ' !

ColorPredicateDetector subclass: #ChromaticColorDetector
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: ''.

  (Smalltalk at: #ChromaticColorDetector) comment: '

	This is an implementation of the skin detection algorithm described
	  in "The Skin Color Based Head Detection"

	URL: http://www.comp.nus.edu.sg/~fanlx/face_detector/skin_color.html

	@1999, Alejandro F. Reimondo aleReimondo@sugarWeb.com
'    !

ColorPredicateDetector subclass: #KjeldsenAndKenderDetector
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: ''.

  (Smalltalk at: #KjeldsenAndKenderDetector) comment: '

	This is an implementation of the skin detection algorithm described
	  in Kjeldsen and Kender "Finding Skin in Color Images."

	The algorithm makes use of a Color Predicate (CP)
	 which is trained by a pre-segmented image.

	@1999, Alejandro F. Reimondo aleReimondo@sugarWeb.com
' !

SkinDetectionObject subclass: #ColorPredicate
  instanceVariableNames: 
    ' minValue maxValue valueContainerClass threshold values '
  classVariableNames: ''
  poolDictionaries: '' !

ColorPredicate subclass: #ChromaticColorPredicate
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: ''   !

ChromaticColorPredicate subclass: #HSColorPredicate
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

ColorPredicate subclass: #HSIColorPredicate
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !


!ColorDetector methods !
 
, aColorDetector
	"Returns aColorDetectorComposite with the receiver and aColorDetector."

	^ColorDetectorComposite
		with: self
		with: aColorDetector!
   
defaultMask
	" Private - Returns the color to be used as mask color."

	^Color
		red: 255
		green: 255
		blue: 255!
   
detectPixel: aPoint with: anImageProcessor
	" Private - Detect color for pixel at aPoint.
	Returns true if the pixel at aPoint success detection. "

	^self implementedBySubclass!
  
detectWith: anImageProcessor
	"Detect using current color predicate using the receiver."

	^self
		detectWith: anImageProcessor
		mask: self defaultMask
					"default mask as default no-skin color mask"!
   
detectWith: anImageProcessor doing: aBlock
	"Detect using current color predicate using the receiver.
	Evaluates aBlock on each detected skin offset. "

	| extent |
	extent := anImageProcessor extent - 1.
	1 to: extent y do: [:y|
		1 to: extent x do: [:x|
			(self detectPixel: x@y with: anImageProcessor)
			ifFalse: [ aBlock value: x@y ].
		].
	].!
   
detectWith: anImageProcessor mask: aColor
	"Detect using current color predicate using the receiver.
	Replace non-skin pixels with mask color."

	^self
		detectWith: anImageProcessor
		doing: [:aPoint| anImageProcessor colorAt: aPoint put: aColor ]!
 
trainWith: anImageProcessor
	"Train the color predicate using the receiver."

	^self
		trainWith: anImageProcessor
		mask: self defaultMask
					"default mask as default no-skin color mask"!

trainWith: anImageProcessor mask: aColor
	"Train the color predicate of the receiver using a pre-segmented image (from anImageProcessor) and mask color aColor."

	| tuple maskRGB extent rgb |
	maskRGB := Array
		with: aColor red
		with: aColor green
		with: aColor blue.
	extent := anImageProcessor extent - 1.
	1 to: extent y do: [:y|
		1 to: extent x do: [:x|
			rgb := anImageProcessor rgbAt: x@y.
			tuple := self tupleFrom: rgb with: anImageProcessor.
			tuple notNil
			ifTrue: [
				maskRGB = rgb
				ifTrue: [ self colorPredicate decrement: tuple ]
				ifFalse: [ self colorPredicate increment: tuple ]
			].
		].
	].!
  
tupleFrom: rgb with: anImageProcessor
	" Private - Returns the trainee tuple for rgb color.
	Returns nil if pixel color is considered invalid for training the receiver.
	This method can be refined by subclasses."

	^rgb! !



!ColorDetectorComposite class methods !
  
with: aColorDetector
	"Returns a non-empty instance of the receiver."

	^self new
		add: aColorDetector;
		yourself!
   
with: aColorDetector
with: bColorDetector
	"Returns a non-empty instance of the receiver."

	^self new
		add: aColorDetector;
		add: bColorDetector;
		yourself!
 
with: aColorDetector
	with: bColorDetector
	with: cColorDetector
	"Returns a non-empty instance of the receiver."

	^self new
		add: aColorDetector;
		add: bColorDetector;
		add: cColorDetector;
		yourself!
 
with: aColorDetector
	with: bColorDetector
	with: cColorDetector
	with: dColorDetector
	"Returns a non-empty instance of the receiver."

	^self new
		add: aColorDetector;
		add: bColorDetector;
		add: cColorDetector;
		add: dColorDetector;
		yourself!
  
withAll: aCollection
	"Returns a non-empty instance of the receiver."

	^self new
		addAll: aCollection;
		yourself! !



!ColorDetectorComposite methods !

, aColorDetector
	"Add aColorDetector to the receiver.
	Returns the receiver."

	self add: aColorDetector.
	^self!
 
add: aColorDetector
	"Add aColorDetector to the receiver."

	^self contents add: aColorDetector!
 
addAll: aCollection
	"Add aCollection to the receiver."

	^self collection addAll: aCollection!
  
contents
    "Returns the instance var of the receiver
     for external manipulation"

	contents isNil ifTrue: [ self initializeContents ].
    ^contents!

detectPixel: aPoint with: anImageProcessor
	" Private - Detect color for pixel at aPoint.
	Returns true if the pixel at aPoint success detection. "

	self contents
		detect: [:one| (one detectPixel: aPoint with: anImageProcessor) not ]
		ifNone: [ ^true ].
	^false!

initializeContents
	" Private - Initializes the receiver's contents"

	contents := OrderedCollection new.!
   
trainWith: anImageProcessor mask: aColor
	"Train the color predicate of the receiver using a pre-segmented image (from anImageProcessor) and mask color aColor."

	self contents do: [:each|
		each trainWith: anImageProcessor mask: aColor
	].! !



!ColorPredicateDetector class methods !
 
clearColorPredicates
    "Clear current color predicates.
	ColorPredicateDetector clearColorPredicates.
	"

	self withAllSubclasses do: [:each| each colorPredicate: nil ].!
   
colorPredicate
    "Returns the instance var of the receiver
     for external manipulation"

	colorPredicate isNil ifTrue: [ self initializeColorPredicate ].
    ^colorPredicate!

colorPredicate: anObject
    "Sets the instance var of the receiver."

    colorPredicate := anObject!
   
initializeColorPredicate
	" Private - Initializes the receiver's colorPredicate. "

	colorPredicate := self colorPredicateClass new.!

with: aColorPredicate
	"Returns an instance of the receiver working with aColorPredicate."

	^self new
		colorPredicate: aColorPredicate;
		yourself! !



!ColorPredicateDetector methods !
   
colorPredicate
    "Returns the instance var of the receiver
     for external manipulation"

	colorPredicate isNil ifTrue: [ self initializeColorPredicate ].
    ^colorPredicate!

colorPredicate: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    colorPredicate := anObject!
   
initializeColorPredicate
	" Private - Initializes the receiver's colorPredicate with default color predicate."

	colorPredicate := self class colorPredicate.! !



!ChromaticColorDetector class methods !
  
colorPredicateClass
	" Private - Returns the receiver's colorPredicate support class."

	^ChromaticColorPredicate!
   
initializeColorPredicate
	" Private - Initializes the receiver's colorPredicate. "

	colorPredicate := self colorPredicateClass forBytes.! !



!ChromaticColorDetector methods !

detectPixel: aPoint with: anImageProcessor
	" Private - Detect color for pixel at aPoint.
	Returns true if the pixel at aPoint success detection. "

	| rgb |
	rgb := anImageProcessor rgbAt: aPoint.
	^self colorPredicate test: rgb! !



!KjeldsenAndKenderDetector class methods !

colorPredicateClass
	" Private - Returns the receiver's colorPredicate support class."

	^HSIColorPredicate! !



!KjeldsenAndKenderDetector methods !
   
detectPixel: aPoint with: anImageProcessor
	" Private - Detect color for pixel at aPoint.
	Returns true if the pixel at aPoint success detection. "

	| rgb hsi |
	rgb := anImageProcessor rgbAt: aPoint.
	^(hsi := anImageProcessor rgb2hsi: rgb) notNil
	and: [ self colorPredicate test: hsi ]!
   
tupleFrom: rgb with: anImageProcessor
	" Private - Returns the trainee tuple for rgb color.
	Returns nil if pixel color is considered invalid for training the receiver.
	This method can be refined by subclasses."

	^anImageProcessor rgb2hsi: rgb! !



!ColorPredicate class methods !

forBytes
	"Create an instance of the receiver for holding values in Byte interval."

	^self new
		minValue: 0;
		maxValue: 255;
		valueContainerClass: ByteArray;
		yourself! !



!ColorPredicate methods !
 
decrement: tuple
	" Decrement the stored value at tuple.
	This method must be implemented by subclasses. "

	^self implementedBySubclass!
   
defaultValue
    " Private - Returns the default values used to initialize values."

	^self threshold!
   
increment: tuple
	" Increment the stored value at tuple.
	This method must be implemented by subclasses. "

	^self implementedBySubclass!
   
initializeMaxValue
	" Private - Initializes the maximum value stored in the values collection."

	maxValue := SmallInteger largestPositive!
  
initializeMinValue
	" Private - Initializes the minimum value stored in the values collection."

    minValue := SmallInteger smallestNegative!
  
initializeThreshold
	" Private - Initializes the receiver's threshold"

	threshold := self zeroValue!

initializeValueContainerClass
	" Private - Initializes the the class to be used to create the receiver's values collection.
    The default action is to use Array, but it can be refined by subclasses."

	valueContainerClass := Array.!
  
maxValue
    "Returns the instance var of the receiver
     for external manipulation"

	maxValue isNil ifTrue: [ self initializeMaxValue ].
    ^maxValue!

maxValue: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    maxValue := anObject!
   
minValue
    "Returns the instance var of the receiver
     for external manipulation"

	minValue isNil ifTrue: [ self initializeMinValue ].
    ^minValue!

minValue: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    minValue := anObject!
   
test: tuple
	" Returns true if the predicate success for matching tuple.
	This method must be implemented by subclasses. "

	^self implementedBySubclass!
   
threshold
    "Returns the instance var of the receiver
     for external manipulation"

	threshold isNil ifTrue: [ self initializeThreshold ].
    ^threshold!

threshold: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    threshold := anObject!
 
valueContainerClass
    "Returns the instance var of the receiver
     for external manipulation"

	valueContainerClass isNil ifTrue: [ self initializeValueContainerClass ].
    ^valueContainerClass!

valueContainerClass: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    valueContainerClass := anObject!
 
zeroValue
    " Private - Returns the default values used to initialize values."

	^self maxValue + self minValue // 2! !



!ChromaticColorPredicate methods !
  
asDIB
	"Build a DIB with the receiver's values."

	| dib bitmap column gray max |
	bitmap := Bitmap
	    width: self values size + 1
		height: self values first size + 1
	    planes: 1 bitCount: 24.
	dib := ProcessableDIB fromDIB: bitmap asDIB.
	bitmap release.
	max := self maximumValue.
	max = 0 ifTrue: [ ^dib ].
	1 to: self values size do: [:x|
		column := self values at: x.
		1 to: column size do: [:y|
			gray := 255 - ((column at: y) / max * 255) truncated.
			dib processor
				rgbAt: x@y
				put: (Array with: gray with: gray with: gray)
		].
	].
	^dib!

decrement: rgb
	" Decrement the stored value at red,green,blue (rgb). "

	| tuple container value |
	tuple := self rgb2tuple: rgb.
	tuple isNil ifTrue: [ ^self ].
	container := self values at: tuple first.
	value := (container at: tuple last) - 1 max: self minValue.
	^container at: tuple last put: value!
   
height
	" Private - Returns the receiver's height. "

	^100!
 
increment: rgb
	" Increment the stored value at red,green,blue (rgb). "

	| tuple container value |
	tuple := self rgb2tuple: rgb.
	tuple isNil ifTrue: [ ^self ].
	container := self values at: tuple first.
	value := (container at: tuple last) + 1 min: self maxValue.
	^container at: tuple last put: value!
   
initializeThreshold
	" Private - Initializes the receiver's threshold"

	| delta zero |
	values isNil ifTrue: [
		^super initializeThreshold
	].
	threshold := self maximumValue - self zeroValue // 100 max: self zeroValue + 17.!
  
initializeValues
	" Private - Initializes the receiver's values. "

	| row |
	values := Array new: self width.
	1 to: self width do: [:i|
		row := self valueContainerClass new: self height.
		row atAllPut: self zeroValue.
		values at: i put: row.
	].!

maximumValue
	" Private - Retursn the maximum value stored in the receiver. "

	| max |
	max := 0.
	self values do: [:column|
		column do: [:value|
			max := max max: value
		]
	].
	^max!
   
rgb2tuple: rgb
	" Private - Returns the (r,g) tuple from rgb values.
	A chromatic color space is converted from the RGB space: 
		r = R / (R + G + B).
		g = G / (R + G + B).
	The features of such (r,g) space: 
		The 2-D (r,g) space is a subspace of 3-D RGB space. 
		The third component b is redundant, since r+g+b=1. 
		Instead of the brightness, the ratio of R,G component to the brightness determine the converted color, which means the brightness component is removed.

	WARNING: Returns anArray of integer values representing the percent of r&g values.
	Return value will be in the 1..100 interval.
	"

	| r g sum |
	sum := (rgb at: 1) + (rgb at: 2) + (rgb at: 3).
	sum = 0 ifTrue: [ ^Array with: 1 with: 1 ].
	^Array
		with: (rgb at: 1) * 99 // sum + 1
		with: (rgb at: 2) * 99 // sum + 1!
  
smooth
	"Smooth the receiver's values."

	| tuple sum |
	tuple := Array new: 2.
	2 to: self width - 1 by: 3 do: [:x|
		2 to: self height - 1 by: 3 do: [:y|
			sum := 0.
			x - 1 to: x + 1 do: [:i|
				y - 1 to: y + 1 do: [:j|
					tuple at: 1 put: i; at: 2 put: j.
					sum := sum + (self valueAt: tuple).
				].
			].
			tuple at: 1 put: x; at: 2 put: y.
			self valueAt: tuple put: sum // 9.
		].
	].!

test: rgb
	" Returns true if the predicate success for matching rgb. "

	^(self valueAt: (self rgb2tuple: rgb)) > self threshold!

valueAt: tuple
	" Returns the value for the tuple. "

	tuple isNil ifTrue: [ ^0 ].
	^(self values at: (tuple at: 1)) at: (tuple at: 2)!
 
valueAt: tuple put: aValue
	" Sets the value for the tuple to aValue. "

	tuple isNil ifTrue: [ ^aValue ].
	^(self values at: (tuple at: 1))
		at: (tuple at: 2)
		put: aValue!
   
values
	"Returns the receiver's values."

	values isNil ifTrue: [ self initializeValues ].
	^values!

width
	" Private - Returns the receiver's width. "

	^100! !



!HSColorPredicate methods !
  
rgb2tuple: rgb
	" Private - Returns the (h,s) tuple from rgb values.
	We will use a (Hue,Saturation) 2D color space.
	Return value will be in the 1..100 interval.
	"

	| intensity hue saturation sumRGB r g b x y o3 t3 arg red green blue result |
	red := rgb at: 1.
	green := rgb at: 2.
	blue := rgb at: 3.
	sumRGB := (red + green + blue) asFloat.

	intensity := sumRGB / 3.
	(intensity between: 10 and: 255) ifFalse: [ ^nil ].

	o3 := 1 / 3 asFloat.
	t3 := 2 / 3 asFloat.

	r := red / sumRGB.
	g := green / sumRGB.
	b := blue / sumRGB.

	x := ((r - o3) squared + (g - o3) squared + (b - o3) squared) sqrt.
	x = 0.0 ifTrue: [ ^nil ].
	y := t3 * (r - o3) - o3 * (g - o3) - o3 * (b - o3).
	arg := (y / (x * t3 sqrt) max: -1) min: 1.

	hue := arg arcCos.
	hue > Float pi ifTrue: [ hue := Float pi * 2 - hue ].

	saturation := 1 - (((r min: g) min: b) * 3).

	^Array
		with: (hue / Float pi * 99 + 1) truncated
		with: (saturation * 99 + 1) truncated! !



!HSIColorPredicate methods !
  
adjust: hsi
	" Private - Adjust hue,saturation,intensity values to indices. "

	^Array
		with: (hsi first * 29 + 1) truncated
		with: ((hsi at: 2) * 99 + 1) truncated
		with: (hsi last min: 99) truncated + 1!
  
atHue: hue saturation: saturation intensity: intensity
	" Private - Returns the stored value at tuple (hue,saturation,intensity). "

	| container |
	container := self intensitiesAtHue: hue saturation: saturation create: false.
	^container isNil
		ifTrue: [ self zeroValue ]
		ifFalse: [ container at: intensity ]!

decrement: hsi
	" Decrement the stored value at hue,saturation,intensity (hsi). "

	| tuple |
	tuple := self adjust: hsi.
	^self
		decrementHue: tuple first
		saturation: (tuple at: 2)
		intensity: tuple last!
   
decrementHue: hue saturation: saturation intensity: intensity
	" Private - Decrement the stored value at tuple (hue,saturation,intensity). "

	| container value |
	container := self intensitiesAtHue: hue saturation: saturation create: true.
	value := ((container at: intensity) - 1
			max: self minValue) min: self maxValue.
	^container at: intensity put: value!
   
hueDimension
	" Private - Returns the dimension for hue scale. "

	^100!
 
increment: hsi
	" Increment the stored value at hue,saturation,intensity (hsi). "

	| tuple |
	tuple := self adjust: hsi.
	^self
		incrementHue: tuple first
		saturation: (tuple at: 2)
		intensity: tuple last!
   
incrementHue: hue saturation: saturation intensity: intensity
	" Private - Increment the stored value at tuple (hue,saturation,intensity). "

	| container value |
	container := self intensitiesAtHue: hue saturation: saturation create: true.
	value := ((container at: intensity) + 1
			max: self minValue) min: self maxValue.
	^container at: intensity put: value!
   
intensitiesAtHue: hue saturation: saturation create: create
	" Private - Returns the stored values container for hue and saturation.
	Or nil if cant be solved/created. "

	| container default |
	values isNil ifTrue: [
		create ifFalse: [ ^nil ].
		values := Array new: self hueDimension.
	].
	container := values at: hue.
	container isNil ifTrue: [
		create ifFalse: [ ^nil ].
		default := Array new: self saturationDimension.
		values at: hue put: default.
		container := default.
	].
	(container at: saturation) isNil ifTrue: [
		create ifFalse: [ ^nil ].
		default := self valueContainerClass new: self intensityDimension.
		default atAllPut: self zeroValue.
		container at: saturation put: default.
		^default
	].
	^container at: saturation!

intensityDimension
	" Private - Returns the dimension for intensity scale. "

	^256!
 
saturationDimension
	" Private - Returns the dimension for saturation scale. "

	^100!
   
test: hsi
	" Returns true if the predicate success for matching (hue,saturation,intensity). "

	| tuple |
	tuple := self adjust: hsi.
	^(self
		atHue: tuple first
		saturation: (tuple at: 2)
		intensity: tuple last)
	 > self threshold!

testHue: hue saturation: saturation intensity: intensity
	" Private - Returns true if the predicate success for matching (hue,saturation,intensity). "

	^(self atHue: hue saturation: saturation intensity: intensity) > self threshold! !
